;	This program is free software; you can redistribute it and/or modify
;	it under the terms of the GNU General Public License as published by
;	the Free Software Foundation; either version 2 of the License, or
;	(at your option) any later version.
;
;	This program is distributed in the hope that it will be useful,
;	but WITHOUT ANY WARRANTY; without even the implied warranty of
;	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;	GNU General Public License for more details.
;
;	You should have received a copy of the GNU General Public License
;	along with this program; if not, write to the Free Software
;	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
BCD_Division:
	;Inputs - HL is the dividend, DE is the divisor
	;Output - HL points to the quotient, DE is the remainder
	;The very first thing, check if the divisor > dividend, and return a remainder if so
	push hl
	push de
	ex de,hl
	call BCD_CheckZero
	jp z,DivisionError
	ex de,hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	ex de,hl
	ld e,(hl)
	inc hl
	ld d,(hl)
	res 7,d
	ex de,hl
	;BC dividend - HL divisor are lengths
	or a
	sbc hl,bc
	jr c,BCD_Division_Normal
	jp nz,BCD_Division_Impossible
	;Same length...check high byte
	pop de
	pop hl
	push hl
	push de
	inc hl
	inc hl
	inc de
	inc de
	ex de,hl
	ld a,(de)
	cp (hl)
	jp c,BCD_Division_Impossible
BCD_Division_Normal:
	pop de
	pop hl
	push hl
	push de
	inc hl
	ld a,(hl)
	and 80h
	ld b,a
	dec hl
	inc de
	ld a,(de)
	and 80h
	dec de
	xor b
	ld (temp5),a			;sign of result
	;We must set up the quotient...same size as the dividend
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	;BC is the length...
	ld hl,(areaFreeStart)
	push hl
	inc hl
	inc hl
	add hl,bc
	ld de,(areaFreeEnd)
	or a
	sbc hl,de
	jp nc,MemoryError
	add hl,de
	ld (areaFreeStart),hl
	pop hl
	;HL points to where to make it...
	ld (hl),c
	inc hl
	ld (hl),b
	inc hl
	push hl
BCD_Division_ZeroFill:
	ld (hl),0
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_ZeroFill
	;The quotient is initialized to zero...save the pointer, to the first byte of the
	;data part
	pop hl
	ld (temp8),hl			;(temp8) is the current byte of the quotient...left to right
	ld hl,0
	ld (temp9),hl			;(temp9) is the number of digits that have been put into the quotient
	;HL, DE are still pushed...
	;We must make a copy of the divisor with it shifted to the right one nibble...
	pop hl		;HL is the divisor
	ld (temp10),hl			;(temp10) is the very first byte of the original divisor
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc hl
	push hl
	;BC is the length
	ld a,(hl)
	and 11110000b
	or a
	jr z,BCD_Division_ZeroDivisorNibble
	inc bc
	ld hl,(areaFreeStart)
	push hl
	inc hl
	inc hl
	add hl,bc
	ld de,(areaFreeEnd)
	or a
	sbc hl,de
	jp nc,MemoryError
	add hl,de
	ld (areaFreeStart),hl
	pop de		;DE is the first byte of the new divisor...
	ld (temp11),de							;(temp11) is the first byte of the shifted divisor
	ld a,c
	ld (de),a
	inc de
	ld a,b
	ld (de),a
	inc de
	dec bc
	pop hl		;Source....
	ldir
	;DE is now on that extra byte
	xor a
	ld (de),a
	;Now we must take this new divisor and shift it a nibble right
	ld hl,(temp11)
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc hl
	;BC is the number of bytes
	;A starts out as 0
	push af
BCD_Division_ShiftDivisor:
	pop af
	rrd
	push af
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_ShiftDivisor
	pop af
	jr BCD_Division_ShiftDivisorDone
BCD_Division_ZeroDivisorNibble:
	ld hl,(areaFreeStart)
	push hl
	inc hl
	inc hl
	add hl,bc
	ld de,(areaFreeEnd)
	or a
	sbc hl,de
	jp nc,MemoryError
	add hl,de
	ld (areaFreeStart),hl
	pop de		;DE is the first byte of the new divisor...
	ld (temp11),de							;(temp11) is the first byte of the shifted divisor
	ld a,c
	ld (de),a
	inc de
	ld a,b
	ld (de),a
	inc de
	pop hl		;Source....
	ldir
	;Now we must take this new divisor and shift it a nibble left
	ld hl,(temp11)
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	add hl,bc
	;BC is the number of bytes
	;A starts out as 0
	xor a
	push af
BCD_Division_ShiftDivisor2:
	pop af
	rld
	push af
	dec hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_ShiftDivisor2
	pop af
	ld de,(temp10)
	ld hl,(temp11)
	ld (temp10),hl
	ld (temp11),de
BCD_Division_ShiftDivisorDone:
	;Now we have two copies of the divisor, one normal in (temp10), and the shifted in (temp11)
	;Now we are ready to start dividing
	;But wait, there's more!
	;We will oscillate between the two divisor copies, start on the shifted if the
	;high nibble of the dividend is zero though
	pop de
	;Finally we have the dividend
	inc de
	inc de
	ld a,(de)
	and 11110000b
	or a
	jr z,BCD_Division_UseShiftDivisor
	ld hl,(temp10)
	jr BCD_Division_AlmostReady
BCD_Division_UseShiftDivisor:
	ld hl,(temp11)
BCD_Division_AlmostReady:
	;Now we have the divisor in HL, the dividend in DE
	;We will subtract the divisor from de, and see how many times we can do this until it
	;carries...when it carries we have to add it back, then write the digit to the quotient
	ld (temp13),de					;original dividend in (temp13)...on data byte 1
BCD_Division_Loop:
	in a,(4)
	and 8h
	jp z,BreakError
	xor a
	ld (temp12),a				;(temp12) is the number of times we've subtracted this
	;DE and HL are assumed to be set up
	push hl
	push de
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	;BC is the length of the divisor
	;DE is the starting point of this dividend
	ex de,hl
	add hl,bc			;HL now is the last byte that will be used
	dec hl
	ex de,hl				;now DE
	ld hl,(temp13)
	dec hl
	ld b,(hl)
	res 7,b
	dec hl
	ld c,(hl)
	inc hl
	add hl,bc
	or a
	sbc hl,de			;check
	jp c,BCD_Division_Done
	pop de
	pop hl
BCD_Division_Digit_Loop:
	in a,(4)
	and 8h
	jp z,BreakError
	push de
   push hl
	ld hl,(temp13)
	ex de,hl
	or a
	sbc hl,de
	ld (temp6),hl			;how many extra math operations to do
	pop hl
	pop de
	push de
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	add hl,bc
	ex de,hl
	add hl,bc
	dec hl
	ex de,hl
	or a		;clear carry flag
	push af
BCD_Division_Subtract_Loop:
	in a,(4)
	and 8h
	jp z,BreakError
	pop af
	ld a,(de)
	sbc a,(hl)
	daa
	ld (de),a
	push af
	dec hl
	dec de
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_Subtract_Loop
	;Now anything else?
	ld bc,(temp6)
	ld a,b
	or c
	jr z,BCD_Division_Subtract_SkipExcess
BCD_Division_Subtract_Loop2:
	pop af
	ld a,(de)
	sbc a,0
	daa
	ld (de),a
	push af
	dec hl
	dec de
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_Subtract_Loop2
BCD_Division_Subtract_SkipExcess:
	;The subtraction is done....now we must check, if it carried or not
	;First though, see if we have a left over result
	;DE is the byte before...check it
	pop af
	jr c,BCD_Division_DigitDone
	;No carry...
	ld a,(temp12)
	inc a			;add to our digit value
	daa
	ld (temp12),a
	pop hl
	pop de			;DE is the dividend...HL divisor
	;Let's have at it again!
	jr BCD_Division_Digit_Loop
BCD_Division_DigitDone:
	;We carried, so we had best add back the divisor
	pop hl
	pop de
	push de
   push hl
	ld hl,(temp13)
	ex de,hl
	or a
	sbc hl,de
	ld (temp6),hl			;how many extra math operations to do
	pop hl
	pop de
	push de
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	add hl,bc
	ex de,hl
	add hl,bc
	dec hl
	ex de,hl
	or a
	push af
BCD_Division_Addition_Loop:
	pop af
	ld a,(de)
	adc a,(hl)
	daa
	ld (de),a
	push af
	dec hl
	dec de
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_Addition_Loop
	ld bc,(temp6)
	ld a,b
	or c
	jr z,BCD_Division_Addition_SkipExcess
BCD_Division_Addition_Loop2:
	pop af
	ld a,(de)
	adc a,0
	daa
	ld (de),a
	push af
	dec hl
	dec de
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_Addition_Loop2
BCD_Division_Addition_SkipExcess:
	pop af
	;Okay...now what?
	ld a,(temp12)			;the digits...
	ld b,a
	ld a,(temp9)
	and 1
	jr z,BCD_Division_StoreDigitsHigh
	ld hl,(temp8)
	ld a,b
	or (hl)
	ld (hl),a
	inc hl
	ld (temp8),hl
	jr BCD_Division_StoreDigitsDone
BCD_Division_StoreDigitsHigh:
	rlc b
	rlc b
	rlc b
	rlc b
	ld hl,(temp8)
	ld (hl),b
BCD_Division_StoreDigitsDone:
	ld hl,(temp9)
	inc hl
	ld (temp9),hl			;we've updated the result...now just update the dividend and divisor pointers
	pop hl
	pop de
	;We have the dividend in DE...divisor in HL
	ld bc,(temp10)
	or a
	sbc hl,bc
	jr z,BCD_Division_WasOriginalDividend
	ld hl,(temp10)
	inc de
	jp BCD_Division_Loop
BCD_Division_WasOriginalDividend:
	ld hl,(temp11)
	jp BCD_Division_Loop
BCD_Division_Done:
	pop de
	pop hl
	;Now we must clean up the quotient...eliminate empty space
	ld hl,(temp8)
	ld bc,(temp9)
	srl b
	rr c
	jr nc,BCD_Division_Even
	;it was odd
	scf
	sbc hl,bc
	inc bc
	ld (hl),b
	dec hl
	ld (hl),c
	inc hl
	inc hl
	ld a,(hl)
	and 11110000b
	or a
	jr nz,BCD_Division_IsOdd
	;We have an extra 0 in front...kill it
	ld bc,(temp9)
	dec bc
	ld (temp9),bc
	dec hl
	ld (hl),b
	dec hl
	ld (hl),c
	inc hl
	inc bc
	add hl,bc
	xor a
	push af
BCD_Division_OddLoop2:
	pop af
	rld
	dec hl
	push af
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_OddLoop2
	pop af
	jr BCD_Division_Even
BCD_Division_IsOdd:
	xor a
	push af
BCD_Division_OddLoop:
	pop af
	rrd
	inc hl
	push af
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_OddLoop
	pop af
	ld bc,(temp9)
	inc bc
	ld (temp9),bc
	ld hl,(temp8)
	inc hl
	ld (temp8),hl
BCD_Division_Even:
	ld bc,(temp9)
	srl b
	rr c
	ld hl,(temp8)
	scf
	sbc hl,bc
	ld (hl),b
	dec hl
	ld (hl),c
	;HL now points to the quotient
	push hl
	;Get the remainder
	ld hl,(temp13)
	dec hl
	ld b,(hl)
	res 7,b
	dec hl
	ld c,(hl)
	inc hl
	inc hl
	;Loop through until we find a nonzero byte
BCD_Division_Done_Loop:
	ld a,(hl)
	or a
	jr nz,BCD_Division_Done_NonZero
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_Division_Done_Loop
	;The whole thing was zeros!
	ld hl,(temp13)
	dec hl
	ld (hl),0
	dec hl
	ld (hl),1
	ex de,hl
	pop hl
	jr BCD_Division_SignOnTheLine
BCD_Division_Done_NonZero:
	dec hl
	ld (hl),b
	dec hl
	ld (hl),c
	ex de,hl
	pop hl
BCD_Division_SignOnTheLine:
	;One more thing...the sign
	inc hl
	inc de
	ld a,(temp5)
	ld b,a
	or (hl)
	ld (hl),a
	dec hl
	ld a,(de)
	or b
	ld (de),a
	dec de
	ret
BCD_Division_Impossible:
	;Return the original dividend as the remainder...quotient is 0
	pop de
	call ReturnZero
	pop de
	ret
DivisionError:
	ld a,eDivZero
	ld sp,(errorStack)
	ld hl,(errorPointer)
	jp (hl)
