;	This program is free software; you can redistribute it and/or modify
;	it under the terms of the GNU General Public License as published by
;	the Free Software Foundation; either version 2 of the License, or
;	(at your option) any later version.
;
;	This program is distributed in the hope that it will be useful,
;	but WITHOUT ANY WARRANTY; without even the implied warranty of
;	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;	GNU General Public License for more details.
;
;	You should have received a copy of the GNU General Public License
;	along with this program; if not, write to the Free Software
;	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


; Prime number functions
IsPrime:
	;Input - HL is a number to check for primality
	;Output - the results of the Rabin-Miller primality test are shown
	ld de,(currow)
	push de
	ld de,7
	ld (currow),de
	call BCD_CheckOne
	jp z,IsPrime_False
	push hl
	;Check for 2, because 2 is prime
	ld a,2
	call ReturnX
	pop de
	call BCD_Compare
	ld b,0
	jp z,IsPrime_True		;it is 2
	push de
	ld a,3
	call ReturnX
	pop de
	call BCD_Compare
	ld b,0
	jp z,IsPrime_True
	ld h,d
	ld l,e
	;Check if it's even...if so, then definitely not prime
	ld c,(hl)
	inc hl
	ld b,(hl)
	bit 7,b
	jp nz,IsPrime_False		;no negative numbers are primes
	add hl,bc
	bit 0,(hl)
	jp z,IsPrime_False
	;Okay...now...we have n...n-1 = 2^d * s
	;let's find d and s
	ld h,d
	ld l,e
	ld (temp14),hl				;temp14 is the original n
	call BCD_Clone
	push hl
	call BCD_Decrement		;n-1
	pop hl						;HL points to n-1
	;To find s and d, let d = 1 and s = (n-1)/2. While s % 1 = 0, { r++; s=s/2; }
	call DivisionBy2
	;HL is the new S
	ld (temp15),hl				;temp15 is s
	call ReturnOne
	ld (temp16),hl				;temp16 is d
	call ReturnOne
	ld (temp17),hl				;temp17 is a temporary pointer to a number 1
IsPrime_FindDSLoop:
	ld hl,(temp15)
	ld c,(hl)
	inc hl
	ld b,(hl)
	add hl,bc
	bit 0,(hl)
	jr nz,IsPrime_DSFound
	;Increment D, divide S by 2
	ld hl,(temp16)
	ld de,(temp17)
	call BCD_Add
	ld (temp16),hl
	ld hl,(temp15)
	call DivisionBy2
	jr IsPrime_FindDSLoop
IsPrime_DSFound:
	;Now clean up S...there might be zeroes at the front
	ld hl,(temp15)
	call ClearZeroes
	ld (temp15),hl
	;Okay now we have our basic setup...let's find the size of the number
	;and determine how many tests to run
	ld de,(temp14)
	ld c,8
	ld hl,IsPrime_Levels
IsPrime_NumberOfTests_Loop:
	push bc
	call BCD_Compare
	jr z,IsPrime_NumberOfTests_Calculate
	jr c,IsPrime_NumberOfTests_Calculate
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
	add hl,bc
	pop bc
	dec c
	jr IsPrime_NumberOfTests_Loop
IsPrime_NumberOfTests_Calculate:
	pop bc
	ld a,c
	ld b,0
	cp 8
	jr nz,IsPrime_NumberOfTests_Okay
	ld bc,1*256+12
IsPrime_NumberOfTests_Okay:
	;Now...run C tests on the number
	ld hl,IsPrime_Bases
IsPrime_Testing_Loop:
	push bc
	ld (errorStack),sp
	ld bc,IsPrime_OnError
	ld (errorPointer),bc
	ld ix,(areaFreeStart)
	push ix
	;Get a base
	push hl
	call IsPrime_Test
	pop hl
	pop ix
	pop bc
	jp nz,IsPrime_False		;it failed
	ld (areaFreeStart),ix
	;It passed the test...keep going
	push bc
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
	add hl,bc
	;HL is the next base
	pop bc
	bit onInterrupt,(iy+onFlags)
	jr nz,IsPrime_True		;interrupted
	push bc
	push hl
	;Show a status indicator - (x) where x is the tests finished so far...
	ld a,12
	sub c
	ld h,0
	ld l,a
	push hl
	ld a,'('
	B_CALL PutC
	pop hl
	call LDispHL
	ld a,')'
	B_CALL PutC
	;Reset cursor position
	xor a
	ld (curcol),a
	pop hl
	;End status indicator
	pop bc
	dec c
	jr nz,IsPrime_Testing_Loop
	;We passed all the tests!
IsPrime_True:
	push bc
	res flagModular,(iy+asm_Flag1)
	B_CALL RunIndicOn
	ld a,busyPause
	ld (indicBusy),a
	res indicOnly,(iy+indicFlags)
	ld hl,IsPrime_Data_True
	call putstr
	pop bc
	;If B=0 then we are completely sure of the answer. If B=1 then show a confidence level
	;Which is in C
	ld a,12
	sub c
	ld c,a
	ld a,b
	or a
	jr nz,IsPrime_True_ShowConfidence
	call putstr
	ld a,'+'
	B_CALL PutC
	call putstr
IsPrime_True_EXITNOW:
	pop de
	ld (currow),de
	res onInterrupt,(iy+onFlags)
	B_CALL GetKey
	ld (errorStack),sp
	ld hl,ErrorHandler
	ld (errorPointer),hl
	jp Interface_Next
IsPrime_True_ShowConfidence:
	call putstr
	push hl
	ld h,0
	ld l,c
	call LDispHL
	pop hl
	call putstr
	jr IsPrime_True_EXITNOW
IsPrime_Bases:
	;2
	DW 1
	DB 02h
	;3
	DW 1
	DB 03h
	;5
	DW 1
	DB 05h
	;7
	DW 1
	DB 07h
	;11
	DW 1
	DB 11h
	;13
	DW 1
	DB 13h
	;17
	DW 1
	DB 17h
	;19
	DW 1
	DB 19h
	;23
	DW 1
	DB 23h
	;29
	DW 1
	DB 29h
	;31
	DW 1
	DB 31h
	;37
	DW 1
	DB 37h
IsPrime_Levels:
	;341550071728321
	DW 8
	DB 03h,41h,55h,00h,71h,72h,83h,21h
	;3474749660383
	DW 7
	DB 03h,47h,47h,49h,66h,03h,83h
	;2152302898747
	DW 7
	DB 02h,15h,23h,02h,89h,87h,47h
	;3215031751
	DW 5
	DB 32h,15h,03h,17h,51h
	;25326001
	DW 0004h
	DB 25h,32h,60h,01h
	;1373653
	DW 0004h
	DB 01h,37h,36h,53h
	;End
	DW 1
	DB 00h
IsPrime_Test:
	;Input - HL points to base a to test n with
	;Output - Z if it passed, NZ if it failed
	;First thing...check if a^s mod n = 1. If so, it passes
	ld (temp18),hl		;temp18 is a
	ld de,(temp14)
	ld (tempN0),de			;tempN0 is the modulus when doing modular exponentation
	set flagModular,(iy+asm_Flag1)
	ld de,(temp15)
	call Exponent
	call BCD_CheckOne
	ret z
	;Okay now for more tests
	push hl
	ld hl,(temp16)
	call BCD_Clone
	ld (temp20),hl				;temp20 is j, counter
	pop hl
	ld (temp19),hl				;temp19 is a^s mod p, this becomes h
IsPrime_Test_Loop:
	;Check if h = n - 1
	ld hl,(temp19)
	call BCD_Clone
	ld de,(temp17)
	call BCD_Add
	ld de,(temp14)
	call BCD_Compare
	ret z			;it passes if h+1 = n
	;Now, h = h^2 mod p
	ld hl,(temp19)
	ld d,h
	ld e,l
	call BCD_Multiply			;squared
	ld de,(temp14)
	call BCD_Division
	ld (temp19),de
	;Now decrement j
	ld hl,(temp20)
	push hl
	call BCD_Decrement
	pop hl
	call BCD_CheckZero
	jr nz,IsPrime_Test_Loop
	or 1
	ret
IsPrime_False:
	res flagModular,(iy+asm_Flag1)
	B_CALL RunIndicOn
	ld a,busyPause
	ld (indicBusy),a
	res indicOnly,(iy+indicFlags)
	ld hl,IsPrime_Data_False
	call putstr
	jp IsPrime_True_EXITNOW
IsPrime_Data_True:
	DB " true",0
	DB " (",0
	DB ")",0
IsPrime_Data_False:
	DB " false",0
IsPrime_OnError:
	pop bc
	jp IsPrime_True
