;	This program is free software; you can redistribute it and/or modify
;	it under the terms of the GNU General Public License as published by
;	the Free Software Foundation; either version 2 of the License, or
;	(at your option) any later version.
;
;	This program is distributed in the hope that it will be useful,
;	but WITHOUT ANY WARRANTY; without even the implied warranty of
;	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;	GNU General Public License for more details.
;
;	You should have received a copy of the GNU General Public License
;	along with this program; if not, write to the Free Software
;	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
Square_Root:
	;Input - HL points to the number (n^2)
	;Output - HL is the integer square root (n) of the number, DE the remainder
	inc hl
	ld a,(hl)
	dec hl
	bit 7,a
	jp nz,SquareRootError	 ;no square roots of negative numbers please
	call BCD_Clone
	ld (temp8),hl			;temp8 is a pointer to the remainder/working number
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc hl
	ld (temp9),hl			;temp9 is the first digit of our n^2 we are dealing with
	ld (temp12),hl			;temp12 is the last digit...
	dec hl
	add hl,bc
	ld (temp10),hl			;temp10 is the last digit of the number...
	;Now make a blank number the same length as n^2, filled with FFh. This is where n will go
	pop hl
	call BCD_Clone
	ld (temp13),hl			;answer pointer
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc hl
	ld (temp11),hl			;temp11 is the pointer to where to store the next digit of the answer
	ld (hl),0FFh
	dec bc
	ld d,h
	ld e,l
	inc de
	ld a,b
	or c
	jr z,Square_Root_SkipInitialize
	ldir
Square_Root_SkipInitialize:
	;Determine by inspection the first square root digit
	ld hl,(temp9)
	ld a,(hl)
	ld hl,Square_Root_Table
	ld c,0
Square_Root_Inspection_Loop:
	sub (hl)
	jr c,Square_Root_Inspection_Complete
	add a,(hl)
	inc c
	inc hl
	jr Square_Root_Inspection_Loop
Square_Root_Inspection_Complete:
	ex de,hl
	;We have the first digit...store it
	ld hl,(temp11)
	ld (hl),c
	inc hl
	ld (temp11),hl
	ld hl,(temp9)
	ld a,(hl)
	dec de
	ex de,hl
	ld c,(hl)
	sub c				;these two lines could have been sub (hl) if the stupid debugger didn't have bugs
	daa
	ld (de),a		;subtracted
	ld de,(temp9)
	ld hl,(temp10)
	or a
	sbc hl,de
	jp z,Square_Root_Done	;if it is a single digit number then we are done
Square_Root_MainLoop:
	bit onInterrupt,(iy+onFlags)
	jp nz,BreakError
	;Now...bring down two more digits for the remainder
	ld hl,(temp9)
	xor a
	cp (hl)
	jr nz,Square_Root_MainLoop_NoExtra
	inc hl
	ld (temp9),hl
Square_Root_MainLoop_NoExtra:
	ld hl,(temp12)
	inc hl
	ld (temp12),hl
	;Now get the square root so far...double it, multiply by 10
	ld hl,(areaFreeStart)
	ld (temp14),hl				;temp14 is where to restore the areaFreeStart pointer at the end of the operations
	ld hl,(temp13)
	call BCD_Clone
	;HL points to the copy...pack it up
	call Square_Root_PackIt
	;HL points to the number, multiply by 10 now
	inc hl
	inc hl
	ld a,(hl)
	cp 5
	jr c,Square_Root_Multiply10_Normal
	push hl
	ld hl,(areaFreeStart)
	inc hl
	ld bc,(areaFreeEnd)
	or a
	sbc hl,bc
	jp nc,MemoryError
	add hl,bc
	ld (areaFreeStart),hl
	pop hl
	dec hl
	dec hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc bc
	ld (hl),b
	dec hl
	ld (hl),c
	;length increased by one
	push hl
	push bc
	dec bc
	inc hl
	add hl,bc
	ld d,h
	ld e,l
	inc de
	lddr			;shift it up
	pop bc
	pop hl
	inc hl
	inc hl
	ld (hl),0
Square_Root_Multiply10_Normal:
	dec hl
	dec hl
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	add hl,bc
	ld e,0
	push af
Square_Root_Multiply10_Loop:
	pop af
	ld a,(hl)
	adc a,a
	daa
	ld (hl),a
	ld a,e
	rld
	ld e,a
	dec hl
	push af
	dec bc
	ld a,b
	or c
	jr nz,Square_Root_Multiply10_Loop
	pop af
	ld hl,(temp8)
	call BCD_Clone
	pop de		;DE is the number to divide the remainder by
	ld (temp17),de		;temp17 is the number
	call Square_Root_Check_Go
	push ix		;save the digit value
	pop bc
	ld a,c
	ld (temp18),a		;temp18 is the digit value
 	;Set rightmost digit to the digit value
	ld hl,(temp17)
	ld c,(hl)
	inc hl
	ld b,(hl)
	add hl,bc
	ld c,a
	ld a,(hl)
	and 11110000b
	or c
	ld (hl),a
	;Now update the cloned remainder
	ld hl,(temp8)
	call BCD_Clone
	push hl
	ld b,h
	ld c,l
	ld hl,(temp12)
	ld de,(temp8)
	or a
	sbc hl,de
	add hl,bc
	ld a,(temp18)
	ld c,a
	ld a,(hl)
	and 11110000b
	or c
	ld (hl),a
	pop hl
	ld de,(temp17)
	call Square_Root_Check_Go
	push ix
	pop bc
	ld a,(temp18)
	cp c
	jp z,Square_Root_Match
	;We need to decrease the digit
	ld hl,(temp17)
	ld c,(hl)
	inc hl
	ld b,(hl)
	add hl,bc
	ld a,(hl)
	jr Square_Root_BackTrack_FirstTime
Square_Root_Backtrack_StackFree:
	pop bc
	pop hl
Square_Root_Backtrack:
	bit onInterrupt,(iy+onFlags)
	jp nz,BreakError
	ld hl,(temp17)
	ld c,(hl)
	inc hl
	ld b,(hl)
	add hl,bc
	ld a,(hl)		;lowered digit...
	dec a
	ld (hl),a
Square_Root_BackTrack_FirstTime:
	;make a separate BCD number for digit
	and 00001111b
	ld (temp18),a
	call ReturnX
	ld de,(temp17)
	call BCD_Multiply
	;HL is the result...compare it to the remainder
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
	push hl
	push bc
	;BC is the size of our answer
	ld hl,(temp8)
	call BCD_Clone
	push hl
	ld (temp0),hl
	push hl
	ld hl,(temp9)
	ld de,(temp8)
	or a
	sbc hl,de
	pop de
	add hl,de
	ld (temp15),hl				;temp15 is the first digit of this clone
	ld hl,(temp12)
	ld de,(temp8)
	or a
	sbc hl,de
	pop de
	add hl,de
	ld (temp16),hl				;temp16 is the last digit to work with
	ld hl,(temp16)
	ld de,(temp15)
	dec de
	or a
	sbc hl,de
	;HL is the length of the remainder to work with
	sbc hl,bc
	jr c,Square_Root_StillLarge
	;Length is okay...check values
	pop bc
	pop hl
	add hl,bc
	dec hl
Square_Root_Backtrack_Outer_Loop:
	push hl
	push bc
	ld de,(temp16)
	xor a
	push af
Square_Root_Backtrack_Loop:
	pop af
	ld a,(de)
	sbc (hl)
	daa
	ld (de),a
	push af
	dec hl
	dec de
	dec bc
	ld a,b
	or c
	jr nz,Square_Root_Backtrack_Loop
	ex de,hl
Square_Root_Backtrack_Finish_Off:
	ld de,(temp15)
	or a
	sbc hl,de
	jr c,Square_Root_Backtrack_Done
	add hl,de
	pop af
	ld a,(hl)
	sbc a,0
	daa
	ld (hl),a
	push af
	dec hl 	
	jr Square_Root_Backtrack_Finish_Off
Square_Root_Backtrack_Done:
	pop af
	jp c,Square_Root_Backtrack_StackFree
	pop bc
	pop hl
	;Done now
	ld a,(temp18)
	ld hl,(temp11)
	ld (hl),a
	inc hl
	ld (temp11),hl
	ld hl,(temp0)
	ld c,(hl)
	inc hl
	ld b,(hl)
	dec hl
	inc bc
	inc bc
	ld de,(temp8)
	ldir
	push af
	jr Square_Root_Match_Done
Square_Root_StillLarge:
	pop bc
	pop hl
	jp Square_Root_Backtrack
Square_Root_Match:
	ld hl,(temp11)
	ld (hl),a
	inc hl
	ld (temp11),hl
	;update remainder
	ld de,(temp17)
	ld hl,(temp8)
	call Square_Root_Check_Go
	;We must now add back
	ld de,(temp12)
	ld hl,(temp17)
	ld c,(hl)
	inc hl
	ld b,(hl)
	add hl,bc
	xor a
	push af
	ex de,hl
Square_Root_Match_AddLoop1:
	pop af
	ld a,(de)
	adc a,(hl)
	daa
	ld (hl),a
	push af
	dec bc
	dec hl
	dec de
	ld a,b
	or c
	jr nz,Square_Root_Match_AddLoop1
	;Now finish
Square_Root_Match_AddLoop2:
	ld de,(temp15)
	or a
	sbc hl,de
	jr c,Square_Root_Match_Done
	add hl,de
	pop af
	ld a,(hl)
	adc a,0
	daa
	ld (hl),a
	push af
	dec hl 	
	jr Square_Root_Match_AddLoop2
Square_Root_Match_Done:
	;we are all done now...
	pop af
	ld hl,(temp14)
	ld (areaFreeStart),hl
	ld hl,(temp10)
	ld de,(temp12)
	or a
	sbc hl,de
	jp nz,Square_Root_MainLoop
Square_Root_Done:
	;Done with the square root...format the answer
	ld hl,(temp13)
	call Square_Root_PackIt
	push hl
	;Now...strip padding zeroes out of the remainder
	ld hl,(temp8)
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
Square_Root_Done_RemainderLoop:
	ld a,(hl)
	or a
	jr nz,Square_Root_Done_RemainderNonZero
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,Square_Root_Done_RemainderLoop
	ld hl,(temp8)
	ld (hl),1
	inc hl
	ld (hl),0
	inc hl
	ld (hl),0
	dec hl
	dec hl
Square_Root_Done_Remainder:
	ex de,hl
	pop hl
	ret
Square_Root_Done_RemainderNonZero:
	dec hl
	ld (hl),b
	dec hl
	ld (hl),c
	jr Square_Root_Done_Remainder
Square_Root_PackIt:
	push hl
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	ld d,b
	ld e,c
	inc hl
Square_Root_PackIt_FF:
	ld a,0FFh
	cp (hl)
	jr z,Square_Root_PackIt_FF_Done
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,Square_Root_PackIt_FF
Square_Root_PackIt_FF_Done:
	ex de,hl
	or a
	sbc hl,bc	;HL is the number of digits there are
	pop de
	push hl
	push de
	ld b,h
	ld c,l
	pop hl
	inc hl
	inc hl
	ld d,h
	ld e,l
	;Now collapse it into packed BCD
	bit 0,c
	jr z,Square_Root_Answer_Loop
	inc hl
	inc de
	dec bc
	ld a,b
	or c
	jr z,Square_Root_Answer_Done
Square_Root_Answer_Loop:
	;extract two digits and store it here
	ld a,(de)
	bit 0,c
	jr z,Square_Root_Answer_Even
	or (hl)
	ld (hl),a
	inc hl
Square_Root_Answer_Go:
	inc de
	dec bc
	ld a,b
	or c
	jr nz,Square_Root_Answer_Loop
Square_Root_Answer_Done:
	;Update the length bytes
	pop bc
	ld a,c
	srl b
	rra
	adc a,0
	ld c,a
	ld a,b
	adc a,0
	ld b,a
	pop hl
	ld (hl),c
	inc hl
	ld (hl),b
	dec hl
	ret
Square_Root_Answer_Even:
	sla a
	sla a
	sla a
	sla a
	ld (hl),a
	jr Square_Root_Answer_Go
Square_Root_Check_Go:
	;DE is the number....to divide the remainder by
	push de
	;HL is the copy...now we must find the right offsets
	push hl
	ld (temp19),hl				;save a pointer
	push hl
	ld hl,(temp9)
	ld de,(temp8)
	or a
	sbc hl,de
	pop de
	add hl,de
	ld (temp15),hl				;temp15 is the first digit of this clone
	ld hl,(temp12)
	ld de,(temp8)
	or a
	sbc hl,de
	pop de
	add hl,de
	ld (temp16),hl				;temp16 is the last digit to work with
	ex de,hl
	pop hl
	;subtract HL from DE...going right of way
	ld c,(hl)
	inc hl
	ld b,(hl)
	add hl,bc
	;HL is now on the last byte...BC is the length to subtract
	ld ix,0		;counter, it's the quotient
Square_Root_Divide_Loop:
	push hl
	push de
	push bc
	xor a
	push af
Square_Root_Divide_SubLoop:
	pop af
	ld a,(de)
	sbc (hl)
	daa
	ld (de),a
	push af
	dec hl
	dec de
	dec bc
	ld a,b
	or c
	jr nz,Square_Root_Divide_SubLoop
Square_Root_Divide_SubLoop2:
	ex de,hl
	ld de,(temp15)
	or a
	sbc hl,de
	jr c,Square_Root_Divide_SubLoopsDone
	add hl,de
	pop af
	ld a,(hl)
	sbc a,0
	daa
	ld (hl),a
	push af
	dec hl
	ex de,hl
	jr Square_Root_Divide_SubLoop2
	;we are all done now...
	pop af
	jr c,Square_Root_Divide_Routine_Done
	push af
Square_Root_Divide_SubLoopsDone:
	pop af
	jr c,Square_Root_Divide_Routine_Done
	inc ix
	pop bc
	pop de
	pop hl
	jr Square_Root_Divide_Loop
Square_Root_Divide_Routine_Done:
	push ix
	pop bc
	ld a,c
	cp 10
	jr c,Square_Root_Divide_Routine_NoAdjust
	dec ix
Square_Root_Divide_Routine_NoAdjust:
	pop bc
	pop de
	pop hl
	ret
Square_Root_Table:
	DB 01h,04h,09h,16h,25h,36h,49h,64h,81h,0FFh
SquareRootError:
	ld a,eDomain
	ld sp,(errorStack)
	ld hl,(errorPointer)
	jp (hl)
