;	This program is free software; you can redistribute it and/or modify
;	it under the terms of the GNU General Public License as published by
;	the Free Software Foundation; either version 2 of the License, or
;	(at your option) any later version.
;
;	This program is distributed in the hope that it will be useful,
;	but WITHOUT ANY WARRANTY; without even the implied warranty of
;	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;	GNU General Public License for more details.
;
;	You should have received a copy of the GNU General Public License
;	along with this program; if not, write to the Free Software
;	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
BCD_Subtract:
	;Inputs - HL and DE point to numbers, DE is subtracted from HL
	;Outputs - HL points to the result
	;
	;First we must do some checking with the signs...
	inc hl
	ld a,(hl)
	dec hl
	and 80h		;bit 7...sign bit. reset for positive, set for negative
	jr nz,BCD_Subtract_HLNegative
	;HL is a positive number....check DE
	inc de
	ld a,(de)
	dec de
	and 80h
	jr z,BCD_Subtract_Start		;if DE is positive too, then we are fine...it'll be handled
	;DE is negative....we will add them instead
	jp BCD_Add_Start
BCD_Subtract_HLNegative:
	;HL is negative...check DE for a course of action
	inc de
	ld a,(de)
	dec de
	and 80h
	jr z,BCD_Subtract_DEPositive
	;HL is negative, and DE is negative
	ex de,hl
	jr BCD_Subtract_Start
BCD_Subtract_DEPositive:
	call BCD_Add_Start
	;Now negate answer
	inc hl
	ld a,80h
	xor (hl)
	ld (hl),a
	dec hl
	ret
BCD_Subtract_Start:
	push de
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	push bc
	push hl
	ld a,(de)
	ld l,a
	inc de
	ld a,(de)
	res 7,a
	ld h,a
	;BC is length of HL...HL is length of DE
	ld (temp3),hl		;length of subtracted number
	ld de,0
	ld (temp2),de		;zeros to pad
	or a
	sbc hl,bc
	jr c,BCD_Subtract_OK
	ld (temp2),hl		;zeros to pad
	add hl,bc
	ld b,h
	ld c,l
BCD_Subtract_OK:
	ld hl,(areaFreeStart)
	ld d,h
	ld e,l
	inc bc
	inc bc
	add hl,bc
	ld bc,(areaFreeEnd)
	or a
	sbc hl,bc
	jp nc,MemoryError
	add hl,bc
	ld (areaFreeStart),hl
	;DE is the address of the result
	ld (temp1),de
	pop hl
	pop bc
	inc hl
	push hl
	ld hl,(temp2)
	add hl,bc
	ld b,h
	ld c,l
	pop hl
	ld a,c
	ld (de),a
	inc de
	ld a,b
	ld (de),a
	inc de
	push bc
	;Adjust DE...pad with zeros if needed
	ld bc,(temp2)
	ld a,b
	or c
	jr z,BCD_Subtract_SkipPad
BCD_Subtract_PadLoop:
	xor a
	ld (de),a
	inc de
	dec bc
	ld a,b
	or c
	jr nz,BCD_Subtract_PadLoop
BCD_Subtract_SkipPad:
	pop bc
	ld (temp0),bc		;total length...of number to subtract from
	push hl
	ld h,b
	ld l,c
	ld bc,(temp2)
	or a
	sbc hl,bc
	ld b,h
	ld c,l
	pop hl
	push bc
	ldir
	;The result is now set up
	pop bc
	dec de
	ex de,hl	;HL now points to the rightmost byte of the result...BC is the length of it
	pop de		;DE now is the first byte of the number to be subtracted
	push bc
	ld a,(de)
	ld c,a
	inc de
	ld a,(de)
	res 7,a
	ld b,a		;BC is the length of DE
	ex de,hl
	add hl,bc
	pop bc
	;HL now is the rightmost byte...DE is the rightmost byte. BC is the length.
	ld bc,(temp3)
	;Now loop through and subtract!
	or a
	push af		;clear carry flag
BCD_Subtract_Loop:
	pop af
	ld a,(de)
	sbc a,(hl)
	daa
	ld (de),a
	push af
	dec hl
	dec de
	dec bc
	ld a,b
	or c
	jr nz,BCD_Subtract_Loop
	;Now do any zero values left
	ld hl,(temp0)
	ld bc,(temp3)
	or a
	sbc hl,bc
	ld a,h
	or l
	jr z,BCD_Subtract_ZeroDone
	ld b,h
	ld c,l
BCD_Subtract_ZeroSub:
	pop af
	ld a,(de)
	sbc a,0
	daa
	ld (de),a
	push af
	dec de
	dec hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_Subtract_ZeroSub
BCD_Subtract_ZeroDone:
	pop af
	;Now check the carry flag...if set, then it's a negative value
	jr nc,BCD_Subtract_Done
	;Negative!
	ld hl,(temp1)
	call BCD_Negate
BCD_Subtract_Done:
	;Now remove any extra zeros...
	ld hl,(temp1)
	ld c,(hl)
	inc hl
	ld b,(hl)
	ld a,b
	and 80h
	ld (temp2),a
	res 7,b
	inc hl
	ld de,0
BCD_Subtract_FindZero:
	ld a,(hl)
	or a
	jr nz,BCD_Subtract_NonZero
	inc de
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_Subtract_FindZero
	;All zero!
	jp ReturnZero
BCD_Subtract_NonZero:
	;DE is the bytes of zero
	;BC is the nonzero bytes
	dec hl
	ld (hl),b
	ld a,(temp2)
	or (hl)
	ld (hl),a
	dec hl
	ld (hl),c
	ret			;all done!
